/*
 *  This is VAX/VMS fast file viewer. Author - S. Kovalyov, INP.
 *
 *  Filetype-sensitive! (distinguishes text/binary files).
 *  Wildcard is allowed!
 */

#include <iodef.h>
#include <rmsdef.h>
#include <lnmdef.h>
#include <jpidef.h>
#include <smgdef.h>
#include <smgmsg.h>
#ifdef __DECC
#  include <smg$routines.h>
#  include <lib$routines.h>
#  include <starlet.h>
#endif
#include <ssdef.h>
#include <fab.h>
#include <nam.h>
#include <rab.h>
#include <xab.h>
#include <descrip.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define UNPRINTABLE '.'

#define VBLOCKLEN    512
#define READBLOCKPOW 5           /* Must be between 0 and 6 */
#if (READBLOCKPOW == 0)
#  define READBLOCKLEN 1         /* multiples of VBLOCKLEN */
#elif (READBLOCKPOW == 1)
#  define READBLOCKLEN 2         /* multiples of VBLOCKLEN */
#elif (READBLOCKPOW == 2)
#  define READBLOCKLEN 4         /* multiples of VBLOCKLEN */
#elif (READBLOCKPOW == 3)
#  define READBLOCKLEN 8         /* multiples of VBLOCKLEN */
#elif (READBLOCKPOW == 4)
#  define READBLOCKLEN 16        /* multiples of VBLOCKLEN */
#elif (READBLOCKPOW == 5)
#  define READBLOCKLEN 32        /* multiples of VBLOCKLEN */
#elif (READBLOCKPOW == 6)
#  define READBLOCKLEN 64        /* multiples of VBLOCKLEN */
#else
#  invalid value assigned to READBLOCKPOW   /* Compile error */
#endif
#define BINRECLEN    16
#define BINTXTLEN    64
#define LONGWORDLEN  4
#define LONGWORDMASK 3
#define FNAMLEN      NAM$C_MAXRSS

#define UP          SMG$K_TRM_UP
#define DOWN        SMG$K_TRM_DOWN
#define RIGHT       SMG$K_TRM_RIGHT
#define LEFT        SMG$K_TRM_LEFT

#define TAB         SMG$K_TRM_CTRLI
#define PREV        SMG$K_TRM_E5
#define NEXT        SMG$K_TRM_E6

#define eNTER       SMG$K_TRM_ENTER
#define ENTER       SMG$K_TRM_CTRLM
#define enTER       SMG$K_TRM_CTRLJ
#define DEL         127
#define hELP        SMG$K_TRM_F15
#define BACKSPACE   SMG$K_TRM_F12
#define ENDOFLINE   SMG$K_TRM_KP2
#define EXIT        SMG$K_TRM_F10
#define BREAK       SMG$K_TRM_CTRLZ
#define FIND        SMG$K_TRM_FIND
#define FIND_CASE   SMG$K_TRM_CTRLF
#define FIND_NEXT   SMG$K_TRM_PF3
#define NEXT_FF     SMG$K_TRM_KP7
#define NEXT_ff     SMG$K_TRM_CTRLL
#define REFRESH     SMG$K_TRM_CTRLW
#define REREAD      SMG$K_TRM_CTRLR

#define TOGGLE_DIR  SMG$K_TRM_F11
#define SET_FORWARD SMG$K_TRM_KP4
#define SET_REVERSE SMG$K_TRM_KP5

#define LNMTBLNAME      "LNM$PROCESS"
#define HLPFLLNM        "LOOK$HELP"
#define EDTFLLNM        "LOOK$EDIT"
#define HELPCOMMAND     "look %slook.help"
#define EDITCOMMAND     "edit %s"
#define DEFAULT_OUTPUT  "sys$output"
#define DEFAULT_INPUT   "sys$input"
#define HELPLINE        "helP eXit Quit Home End scrRight scrLeft Find findNxt eD Go Spawn chWid Txt/hex "
#define INVITE_FIND     "Find : "
#define INVITE_LINE     "Line number or percentage : "
#define NOTHING_FOUND   "Nothing found.  PRESS ANY KEY  "
#define WORKING_FORMAT  " %3d%% processing... (Press \'X\' to cancel) "
#define PLEASE_WAIT     " Wait please... "
#define HEADLINE_FORMAT "File: %s  Row: %5d Col: %3d (%3d%%)  Dir: %s"

#define $_ABNORMAL      -1
#define EXIT_NORMAL     2
#define NO_CLOSE        3
#define RELOOK          4

#define SCREENVER       20
#define SCREENMAX       132
#define SCREENMIN       80
#define SCREENCHNG      13
#define BORDERFIELD     0
#define LASTLINE        24
#define MAXLINE         256
#define BUFDEPTH        20
#define BUFQ            2
#define TOFIND          0
#define TOLINE          1
#define WORKING_STEP    100
#define READKEY_STEP    99
#define FORWARD         0
#define REVERSE         1

#define CHECKRESULT(x)  { if ((x) != SS$_NORMAL) return $_ABNORMAL; }
#define FBEG(y)         { if (sh_seek(vf_inf.home[y]) == -1) return $_ABNORMAL;}
#define BEEP            CHECKRESULT(smg$ring_bell(&vdisp.id))
#define TOLOWER(c)      ((c) >= 'A' && (c) <= 'Z' ? (c) | 0x20:(c))
#define TOUPPER(c)      ((c) >= 'a' && (c) <= 'z' ? (c) & 0x5F:(c))


struct
{
    int bintxt;
    unsigned int *home;
    unsigned int current, end, all;
    struct
    {
        unsigned int pos;
        int beg;
        unsigned int rend;
        char sline[SCREENMAX + 1];
    } head, inv;
} vf_inf =
{   0,
    NULL,
    0, 0, 0,
    { SMG$K_TOP, 1, SMG$M_REVERSE, {'\0'} },
    { SMG$K_BOTTOM, 1, SMG$M_REVERSE, {'\0'} }
};

struct
{
    unsigned long id;
    long          firstrow, firstcol, len, wid;
    struct {long  firstcol, firstrow, len, curcol;} vport;
    struct {long  row, col;} cursor;
    char *f_spec;
} vdisp =
{
    0, 2, 1, MAXLINE, SCREENVER,
    { 1, 1, 0, 1 }, { 1, 1 },
    { '\0' }
};

struct
{
    unsigned long id;
    long firstrow, firstcol, len, wid, rend;
} k_bar =
{
    0, LASTLINE, 1, SCREENMAX, 1, SMG$M_REVERSE
};

char    line[MAXLINE + 1], no_line[MAXLINE + 1];
int     line_len;
char    out_line[8*MAXLINE + 1];
char    search_line [SCREENMAX + 1];
int     search_len = 0;
int     case_sensitive_search = FALSE;
$DESCRIPTOR(outputline, out_line);
$DESCRIPTOR(outstr, INVITE_FIND);
$DESCRIPTOR(tablename, LNMTBLNAME);
$DESCRIPTOR(hlnm, HLPFLLNM);
$DESCRIPTOR(elnm, EDTFLLNM);
int wn = 0;
int old_percent = -1;
int find_col = 0;
unsigned int       rpt = 0;
unsigned short int lnmlen = 0;
unsigned int       p_id = 0, k_id = 0;
unsigned int       screensize = 0;
unsigned int       direction = FORWARD;

struct
{
    unsigned short int  item_length;
    unsigned short int  item_code;
    char               *item_address;
    unsigned short int *item_length_address;
} item_list [2] =
{
    {MAXLINE, LNM$_STRING, line, &lnmlen},
    {0,0,0,0}
};

struct {
    int  flen;
    char fline[SCREENMAX + 1];
} linebuf[BUFQ][BUFDEPTH + 1];

int     fbc[BUFQ] = { 1, 1 };

unsigned int timeout    = 0;
unsigned int rend_find  = SMG$M_BOLD;
unsigned int c_off      = SMG$M_CURSOR_OFF;
unsigned int c_on       = SMG$M_CURSOR_ON;
unsigned int pscroll    = SMG$M_DOWN;
unsigned int scrright   = SMG$M_RIGHT;
unsigned int scrleft    = SMG$M_LEFT;


    /* RMS file I/O routines */

struct FAB fab_of_name, fab_of_file;
struct NAM nam_file;
struct RAB several_bytes;
struct XABFHC file_header;

int inner_ptr = READBLOCKLEN * VBLOCKLEN;
char v_block[READBLOCKLEN * VBLOCKLEN];

char res_name[FNAMLEN];
char rnm[FNAMLEN];
char outname[FNAMLEN];

int asy_stop = 0;

int v_rec_len   = 0;
int align_byte  = 0;


        /* Asynchronious operation completion indication */
void
ast_stop(void)
{
    asy_stop = 1;
}


        /* Expand file name */
char *
sh_fname(char *name_of_file)
{
    if (name_of_file != NULL)
    {
        nam_file.nam$b_bid = NAM$C_BID;
        nam_file.nam$b_bln = NAM$C_BLN;
        nam_file.nam$b_ess = FNAMLEN;
        nam_file.nam$b_esl = FNAMLEN;
        nam_file.nam$l_esa = res_name;
        nam_file.nam$b_nop = 0;
        nam_file.nam$b_rss = FNAMLEN;
        nam_file.nam$b_rsl = FNAMLEN;
        nam_file.nam$l_rsa = rnm;

        fab_of_name.fab$b_bid = FAB$C_BID;
        fab_of_name.fab$b_bln = FAB$C_BLN;
        fab_of_name.fab$l_nam = &nam_file;
        fab_of_name.fab$l_fna = name_of_file;
        fab_of_name.fab$b_fns = strlen(name_of_file);
        fab_of_name.fab$l_fop = FAB$M_OFP;

        fab_of_name.fab$l_sts = 0;
        asy_stop = 0;
        sys$parse(&fab_of_name, ast_stop, ast_stop);
        while(asy_stop == 0);
        if (fab_of_name.fab$l_sts != RMS$_NORMAL) return NULL;
    }

    fab_of_name.fab$w_ifi = 0;
    fab_of_name.fab$l_sts = 0;
    asy_stop = 0;
    sys$search(&fab_of_name, ast_stop, ast_stop);
    while(asy_stop == 0);
    if (fab_of_name.fab$l_sts != RMS$_NORMAL) return NULL;
    strncpy(outname, rnm, nam_file.nam$b_rsl);
    outname[nam_file.nam$b_rsl] = '\0';
    return outname;
}


        /* File name only */
char *
sh_onlyname(void)
{
    return &outname[(int)(nam_file.nam$l_name - nam_file.nam$l_rsa)];
}


        /* Open file */
int
sh_open(char *name_of_file)
{
    if (name_of_file == NULL) return -1;

    inner_ptr = READBLOCKLEN * VBLOCKLEN;

    fab_of_file.fab$b_bid = FAB$C_BID;
    fab_of_file.fab$b_bln = FAB$C_BLN;
    fab_of_file.fab$w_ifi = 0;
    fab_of_file.fab$b_fac = FAB$M_GET | FAB$M_BIO;
    fab_of_file.fab$l_fna = name_of_file;
    fab_of_file.fab$b_fns = strlen(name_of_file);
    fab_of_file.fab$l_fop = FAB$M_SQO | FAB$M_CIF;
    fab_of_file.fab$b_shr = FAB$M_SHRGET | FAB$M_SHRPUT | FAB$M_UPI;
    fab_of_file.fab$b_rtv = 0;
    fab_of_file.fab$l_xab = (void *) &file_header;

    file_header.xab$b_cod = XAB$C_FHC;
    file_header.xab$b_bln = XAB$C_FHCLEN;
    file_header.xab$l_nxt = 0;

    asy_stop = 0;
    fab_of_file.fab$l_sts = 0;
    sys$open(&fab_of_file, ast_stop, ast_stop);
    while(asy_stop == 0);
    if (fab_of_file.fab$l_sts != RMS$_NORMAL) return -1;

    several_bytes.rab$b_bid = RAB$C_BID;
    several_bytes.rab$b_bln = RAB$C_BLN;
    several_bytes.rab$w_isi = 0;
    several_bytes.rab$l_rop = RAB$M_RAH | RAB$M_BIO | RAB$M_NLK | RAB$M_RRL;
    several_bytes.rab$l_fab = &fab_of_file;
    several_bytes.rab$l_xab = 0;

    if (sys$connect(&several_bytes) != RMS$_NORMAL) return -1;

    several_bytes.rab$l_bkt = 0;
    several_bytes.rab$w_usz = (unsigned short)(READBLOCKLEN * VBLOCKLEN);
    several_bytes.rab$l_ubf = v_block;

    v_rec_len   = 0;
    return 0;
}


        /* Read some bytes */
int
sh_read(char *buf, int buflen)
{
    int rest_blk = READBLOCKLEN * VBLOCKLEN - inner_ptr;
    int actual_rest = buflen;
    int status;


    if ((buflen == 1) && (rest_blk > 0))
    {
       *buf = *(several_bytes.rab$l_ubf + inner_ptr);
       inner_ptr++;
       return 1;
    }

    while (actual_rest > rest_blk)
    {
        if (rest_blk > 0)
            memcpy(buf + (buflen - actual_rest),
                   several_bytes.rab$l_ubf + inner_ptr, (size_t)rest_blk);

        /* Next virtual block */
        if (several_bytes.rab$l_bkt == 0)
           several_bytes.rab$l_bkt = 1;
        else
           several_bytes.rab$l_bkt += READBLOCKLEN;

        status = sys$read(&several_bytes);
        if ((status != RMS$_NORMAL) && (status != RMS$_OK_RRL))
            return buflen - actual_rest;

        actual_rest -= rest_blk;
        rest_blk = READBLOCKLEN * VBLOCKLEN;
        inner_ptr = 0;
    }
    if (actual_rest == 0) return buflen;
    memcpy(buf + (buflen - actual_rest),
           several_bytes.rab$l_ubf + inner_ptr, (size_t)actual_rest);
    inner_ptr += actual_rest;
    return buflen;
}


        /* Calculate byte offset */
unsigned int
sh_tell(void)
{
   if (several_bytes.rab$l_bkt == 0) return 0;

   return (unsigned int)(several_bytes.rab$l_bkt * VBLOCKLEN
          + inner_ptr - VBLOCKLEN);
}


        /* End of file offset */
unsigned int
sh_eof(void)
{
    return VBLOCKLEN * (file_header.xab$l_ebk - 1) + file_header.xab$w_ffb;
}


        /* Find position under offset */
int
sh_seek(unsigned int offset)
{
    int status;

#if (VBLOCKLEN == 512)
    if ( (several_bytes.rab$l_bkt == 0) ||
         ( ((several_bytes.rab$l_bkt - 1) >> READBLOCKPOW) !=
           (offset >> (READBLOCKPOW + 9)) ) )
    {
        several_bytes.rab$l_bkt =
                          ((offset >> (READBLOCKPOW + 9)) << READBLOCKPOW) + 1;
        status = sys$read(&several_bytes);
        if ((status != RMS$_NORMAL) && (status != RMS$_OK_RRL)) return -1;
    }
    inner_ptr = offset & (0x7FFF >> (6 - READBLOCKPOW));
#else
    if ( (several_bytes.rab$l_bkt == 0) ||
         ( ((several_bytes.rab$l_bkt - 1) / READBLOCKLEN !=
            offset / (READBLOCKLEN * VBLOCKLEN)) )
    {
        several_bytes.rab$l_bkt = (offset / (READBLOCKLEN * VBLOCKLEN)) *
                                  READBLOCKLEN + 1;
        status = sys$read(&several_bytes);
        if ((status != RMS$_NORMAL) && (status != RMS$_OK_RRL)) return -1;
    }
    inner_ptr = offset % (READBLOCKLEN * VBLOCKLEN);
#endif
    return 0;
}


        /* Retrieve binary line with given length */
int
bin_fgets(char *str, int charnum, int mode)
{
    int i = 0;
    int txt_offset = charnum * 2 + LONGWORDLEN;
    int str_len = txt_offset + LONGWORDLEN + charnum + 1;
    char ch = '\0';
    unsigned char ch2;
    char hex_ch[3];
    char *hex_ptr, *txt_ptr;

    if (mode != SS$_NORMAL)
    {
        if  (sh_read(str, charnum) != charnum) return -1;
        return 0;
    }
    memset(str, ' ', (size_t)str_len);
    hex_ptr = str + txt_offset;
    txt_ptr = hex_ptr + LONGWORDLEN;
    for (i = 0; i < BINRECLEN; i++)
    {
        if  (sh_read(&ch, 1) != 1) return -1;

        ch2 = (unsigned char) ch;
        if (ch2 >= 0xA0)
           hex_ch[0] = (ch2 >> 4) + ('A' - 10);
        else
           hex_ch[0] = (ch2 >> 4) + '0';
        ch2 = ch2 & 0xF;
        if (ch2 >= 0xA)
           hex_ch[1] = ch2 + ('A' - 10);
        else
           hex_ch[1] = ch2 + '0';

        if ((i & LONGWORDMASK) == 0) hex_ptr--;
        *hex_ptr = hex_ch[1];
        hex_ptr--;
        *hex_ptr = hex_ch[0];
        hex_ptr--;
        *txt_ptr = (isprint(ch) == 0) ? UNPRINTABLE : ch;
        txt_ptr++;
    }
    *txt_ptr = '\0';
    return 0;
}


        /* Retrieve text line with given length */
int
txt_fgets(char *str, int charnum, int mode)
{
    int p = 0;

    if (sh_read(str, charnum) != charnum) return -1;
    if (mode == SS$_NORMAL)
    {
        for (p = 0; p < charnum; p++)
            if (isprint(str[p]) == 0) str[p] = '.';
        str[p] = '\0';      /* Terminator */
    }
    return 0;
}


        /* Retrieve a line considering RMS file organization */
int
sh_fgets(char *str, int maxchar, int mode)
{
    int i = 0;
    char ctrl_byte = '\0';
    char *rest_rec = NULL;

    line_len = 0;
    if (maxchar <= 0) return -1;
    if (sh_tell() >= sh_eof()) return -1;
    if (vf_inf.bintxt == $_ABNORMAL)
    {
        line_len = 56;
        return bin_fgets(str, BINRECLEN, mode);
    }
    switch(fab_of_file.fab$b_rfm)
    {
    case FAB$C_VAR:                 /* Usual VMS text files */
    case FAB$C_VFC:                 /* Printer format files */

        if (vf_inf.bintxt == 0) vf_inf.bintxt = SS$_NORMAL;

        if (v_rec_len == 0)
        {
            /* Retrieve record length */

            if (sh_read(&ctrl_byte, 1) != 1) return -1;
            v_rec_len = (int)((unsigned char)ctrl_byte);
            if (sh_read(&ctrl_byte, 1) != 1)
            {
                v_rec_len = 0;
                return -1;
            }
            v_rec_len += ((int)((unsigned char)ctrl_byte) << 8);
            if (v_rec_len == 0)
            {
                *str = '\0';
                return 0;
            }
            if (v_rec_len == 0xFFFF)
            {
                v_rec_len = 0;
                return sh_fgets(str, maxchar, mode);
            }

        /* Retrieve record */

            align_byte = v_rec_len % 2;

            switch(fab_of_file.fab$b_rat)
            {
            case FAB$M_PRN:         /* Printer file */
                if (sh_read(&ctrl_byte, 1) != 1)
                {
                    v_rec_len = 0;
                    return -1;
                }
                v_rec_len--;
                if (sh_read(&ctrl_byte, 1) != 1)
                {
                    v_rec_len = 0;
                    return -1;
                }
                v_rec_len--;
            case FAB$M_FTN:         /* Fortran file */
            case FAB$M_CR:          /* Text file */
            default:                /* Abstract variable-length file */
                break;
            }
        }

        if (maxchar >= v_rec_len)
        {
            if (sh_read(str, v_rec_len) != v_rec_len)
            {
                v_rec_len = 0;
                return -1;
            }
            if (align_byte == 1)
            {
                if (sh_read(&ctrl_byte, 1) != 1)
                {
                    v_rec_len = 0;
                    return -1;
                }
            }
            if (mode == SS$_NORMAL)
            {
                str[v_rec_len] = '\0';
                line_len = v_rec_len;
            }
            v_rec_len = 0;
            return 0;
        }
        else
        {
            if (sh_read(str, maxchar) != maxchar)
            {
                v_rec_len = 0;
                return -1;
            }
            v_rec_len -= maxchar;
            if (mode == SS$_NORMAL)
            {
                str[maxchar] = '\0';
                line_len = maxchar;
            }
            return 1;
        }
    case FAB$C_STMLF:                   /* Files created by C language */
        switch(fab_of_file.fab$b_rat)
        {
        case FAB$M_PRN:
            if (sh_read(&ctrl_byte, 1) != 1) return -1;
            if (sh_read(&ctrl_byte, 1) != 1) return -1;
        case FAB$M_FTN:
        case FAB$M_CR:
            if (vf_inf.bintxt == 0) vf_inf.bintxt = SS$_NORMAL;
            for (v_rec_len = 0; v_rec_len < maxchar; v_rec_len++)
            {
                if (sh_tell() >= sh_eof()) break;
                if (sh_read(&ctrl_byte, 1) != 1) return -1;
                if (ctrl_byte == '\n')
                {
                    if (mode == SS$_NORMAL) str[v_rec_len] = '\0';
                    line_len = v_rec_len;
                    return 0;
                }
                if (mode == SS$_NORMAL)
                    str[v_rec_len] = ctrl_byte;
            }
            if (mode == SS$_NORMAL) str[v_rec_len] = '\0';
            line_len = v_rec_len;
            return 0;
        default:
            if (vf_inf.bintxt == 0)
            {
                vf_inf.bintxt = $_ABNORMAL;
                line_len = 56;
                return bin_fgets(str, BINRECLEN, mode);
            }
            line_len = BINTXTLEN;
            return txt_fgets(str, BINTXTLEN, mode);
        }
    default:                        /* Any other files are binary */
        if (vf_inf.bintxt == 0)
        {
            vf_inf.bintxt = $_ABNORMAL;
            line_len = 56;
            return bin_fgets(str, BINRECLEN, mode);
        }
        line_len = BINTXTLEN;
        return txt_fgets(str, BINTXTLEN, mode);
    }

    /* Should never reach here. */
    return $_ABNORMAL;
}


        /* Close a file */
int
sh_close(void)
{
    asy_stop = 0;
    fab_of_file.fab$l_sts = 0;
    sys$close(&fab_of_file, ast_stop, ast_stop);
    while(asy_stop == 0);
    return (fab_of_file.fab$l_sts == RMS$_NORMAL) ? 0 : -1;
}


    /* Looker routines */


        /* Fill in a line descriptor */
struct dsc$descriptor_s *
outline(char *str, int str_len)
{
    int i, j, k, space_count;

    if (str_len < 0) str_len = strlen (str);
    j = 0;
    for (i = 0; i < str_len; i++) {
        if (str[i] != '\t') {
            /* Copy non-tabs */
            outputline.dsc$a_pointer [j] = str[i];
            j++;
        }
        else {
            /* Replace tabs with spaces */
            space_count = 8 - (j & 7);
            for (k = 0; k < space_count; k++) {
                outputline.dsc$a_pointer [j] = ' ';
                j++;
            }
        }
    }
    outputline.dsc$w_length = j;
    return &outputline;
}


        /* Initialize a screen */
int
init_screen(void)
{
    int j = 0;

    vdisp.vport.firstcol = vdisp.vport.firstrow = vdisp.vport.curcol =
        vdisp.cursor.row = vdisp.cursor.col = 1;

    CHECKRESULT(smg$create_virtual_keyboard(&k_id));

    CHECKRESULT(smg$create_pasteboard(&p_id, outline(DEFAULT_OUTPUT, -1),
        &k_bar.firstrow, &screensize));
    if ((screensize < 10) || (k_bar.firstrow < 5)) return $_ABNORMAL;
    if (screensize > SCREENMAX) screensize = SCREENMAX;
    vdisp.vport.len = screensize - BORDERFIELD;
    vdisp.wid = k_bar.firstrow - 4;

    CHECKRESULT(smg$create_virtual_display(&k_bar.wid,
        &k_bar.len, &k_bar.id));
    CHECKRESULT(smg$put_line(&k_bar.id, outline(HELPLINE,-1), &j, &k_bar.rend));

    CHECKRESULT(smg$create_virtual_display(&vdisp.wid,
        &vdisp.len, &vdisp.id, &SMG$M_DISPLAY_CONTROLS));
    CHECKRESULT(smg$create_viewport(&vdisp.id,
        &vdisp.vport.firstrow, &vdisp.vport.firstcol,
        &vdisp.wid, &vdisp.vport.len));
    CHECKRESULT(smg$home_cursor(&vdisp.id));

    return SS$_NORMAL;
}


        /* Free a screen resources */
int
free_screen(void)
{
    CHECKRESULT(smg$delete_virtual_keyboard(&k_id));
/*
    CHECKRESULT(smg$delete_viewport(&vdisp.id));
    CHECKRESULT(smg$delete_virtual_display(&k_bar.id));
    CHECKRESULT(smg$delete_virtual_display(&vdisp.id));
*/
    CHECKRESULT(smg$delete_pasteboard(&p_id));
    free(vf_inf.home);
    if (sh_close() == -1) return NO_CLOSE;
    return SS$_NORMAL;
}


        /* Fill in a screen page */
int
new_page(void)
{
    int k = 0;

    FBEG(vf_inf.current);
    smg$begin_display_update(&vdisp.id);
    CHECKRESULT(smg$erase_display(&vdisp.id));
    CHECKRESULT(smg$home_cursor(&vdisp.id));
    for (k = 0; (k < vdisp.wid) && (sh_fgets(line, MAXLINE, SS$_NORMAL) != -1);
        k++)
        CHECKRESULT(smg$put_line(&vdisp.id, outline(line, line_len)));
    smg$end_display_update(&vdisp.id);
    return SS$_NORMAL;
}


        /* Put a work indicator */
int
working(void)
{
    unsigned short stop_key;
    int            percent;

    if (wn == 0) CHECKRESULT(smg$set_cursor_mode(&p_id, &c_off));

    if ((wn >= WORKING_STEP) || (wn == 0))
    {
        wn = 0;
        percent = (sh_eof() != 0) ? ((sh_tell() * 100)/sh_eof()) : 100;
        if (percent != old_percent)
        {
            old_percent = percent;

            sprintf(vf_inf.inv.sline, WORKING_FORMAT, percent);
            CHECKRESULT(smg$label_border(&vdisp.id,outline(vf_inf.inv.sline,-1),
                &vf_inf.inv.pos, &vf_inf.inv.beg, &vf_inf.inv.rend));

            stop_key = 0;
            if ( (smg$read_keystroke(&k_id, &stop_key, NULL, &timeout) ==
                  SS$_NORMAL) &&
                 ( (stop_key == EXIT) || (stop_key == BREAK) ||
                   (stop_key == 'X')  || (stop_key == 'x') ) )
            {
                rpt = 0;
                CHECKRESULT(smg$set_cursor_mode(&p_id, &c_on));
                return 0;
            }
        }
    }

    wn++;
    return SS$_NORMAL;
}



        /* Headline a screen page & set screen cursor */
int
headline(void)
{
    int n = vf_inf.current + vdisp.cursor.row;

    CHECKRESULT(smg$set_cursor_abs(&vdisp.id,
        &vdisp.cursor.row, &vdisp.cursor.col));
    sprintf(vf_inf.head.sline, HEADLINE_FORMAT,
        sh_onlyname(), n, vdisp.cursor.col,
        (sh_eof() == 0) ? 100 :
            (n >= vf_inf.all-1) ? 100 : ((vf_inf.home[n-1] * 100)/sh_eof()),
        (direction == FORWARD) ? "Forward" : "Reverse");
    CHECKRESULT(smg$label_border(&vdisp.id, outline(vf_inf.head.sline, -1),
        &vf_inf.head.pos, &vf_inf.head.beg, &vf_inf.head.rend));
    return SS$_NORMAL;
}


        /* Accept a line to buffer. Edition allowed */
int
get_line(int bufnum)
{
    char *invite[] =
    {
        INVITE_FIND, INVITE_LINE
    };
    unsigned short sch = 0;
    int p = 0, q = 0, r = 0;
    int p_rows = vdisp.wid + 2;
    int p_cols = vf_inf.inv.pos + strlen(invite[bufnum]);


    if (linebuf[bufnum][fbc[bufnum]].flen != 0)
    {
        if (fbc[bufnum] == BUFDEPTH)
        {
            for (q = 1; q < BUFDEPTH; q++)
            {
                strcpy(linebuf[bufnum][q].fline, linebuf[bufnum][q + 1].fline);
                linebuf[bufnum][q].flen = linebuf[bufnum][q + 1].flen;
            }
            memset(linebuf[bufnum][fbc[bufnum]].fline, '\0', (size_t)SCREENMAX);
            linebuf[bufnum][fbc[bufnum]].flen = 0;
        }
        else fbc[bufnum]++;
        linebuf[bufnum][fbc[bufnum]].flen = 0;
    }
    r = fbc[bufnum];
    CHECKRESULT(smg$label_border(&vdisp.id, outline(invite[bufnum], -1),
        &vf_inf.inv.pos, &vf_inf.inv.beg, &vf_inf.inv.rend));
    CHECKRESULT(smg$set_physical_cursor(&p_id, &p_rows, &p_cols));
    while ((sch != ENTER) && (sch != eNTER) && (p_cols < vdisp.vport.len))
    if (smg$read_keystroke(&k_id, &sch) == SS$_NORMAL)
    {
        if ((sch < (unsigned short)256) && (isprint(sch) != 0))
        {
            p_cols++;
            linebuf[bufnum][fbc[bufnum]].flen++;
            for (q = linebuf[bufnum][fbc[bufnum]].flen; q > p; q--)
                linebuf[bufnum][fbc[bufnum]].fline[q] =
                    linebuf[bufnum][fbc[bufnum]].fline[q-1];
            linebuf[bufnum][fbc[bufnum]].fline[p] = (char)sch;
            p++;
        }
        else
        {
            switch(sch)
            {
            case RIGHT:
                if (p < linebuf[bufnum][fbc[bufnum]].flen)
                {
                    p_cols++;
                    p++;
                }
                break;
            case LEFT:
                if (p > 0)
                {
                    p_cols--;
                    p--;
                }
                break;
            case SMG$K_TRM_CTRLH:
                p_cols = vf_inf.inv.pos + strlen(invite[bufnum]);
                p = 0;
                break;
            case SMG$K_TRM_CTRLE:
                p_cols = vf_inf.inv.pos + strlen(vf_inf.inv.sline);
                p = linebuf[bufnum][fbc[bufnum]].flen;
                break;
            case UP:
                if (r >= 0)
                {
                    if (r > 0) r--;
                    strcpy(linebuf[bufnum][fbc[bufnum]].fline,
                        linebuf[bufnum][r].fline);
                    linebuf[bufnum][fbc[bufnum]].flen = linebuf[bufnum][r].flen;
                    p_cols -= p;
                    p = 0;
                }
                break;
            case DOWN:
                if (r < fbc[bufnum])
                {
                    r++;
                    if (r == fbc[bufnum]) r = 0;
                    strcpy(linebuf[bufnum][fbc[bufnum]].fline,
                        linebuf[bufnum][r].fline);
                    linebuf[bufnum][fbc[bufnum]].flen = linebuf[bufnum][r].flen;
                    p_cols -= p;
                    p = 0;
                    if (r == 0) r = fbc[bufnum];
                }
                break;
            case (unsigned short)DEL:
                if (p > 0)
                {
                    for (q = p; q < linebuf[bufnum][fbc[bufnum]].flen; q++)
                        linebuf[bufnum][fbc[bufnum]].fline[q-1] =
                            linebuf[bufnum][fbc[bufnum]].fline[q];
                    linebuf[bufnum][fbc[bufnum]].fline
                        [--linebuf[bufnum][fbc[bufnum]].flen] = '\0';
                    p_cols--;
                    p--;
                }
                break;
            case SMG$K_TRM_F13:
                if (p == 0) break;
                sch = (unsigned short)linebuf[bufnum][fbc[bufnum]].fline[p-1];
                for (q = p; q < linebuf[bufnum][fbc[bufnum]].flen; q++)
                    linebuf[bufnum][fbc[bufnum]].fline[q-1] =
                        linebuf[bufnum][fbc[bufnum]].fline[q];
                linebuf[bufnum][fbc[bufnum]].fline
                    [--linebuf[bufnum][fbc[bufnum]].flen] = '\0';
                p_cols--;
                p--;
                if (isalpha((int)sch) != 0)
                    while((p > 0) &&
                        (isalpha((int)linebuf[bufnum][fbc[bufnum]].fline[p-1]) != 0))
                    {
                        for (q = p; q < linebuf[bufnum][fbc[bufnum]].flen; q++)
                            linebuf[bufnum][fbc[bufnum]].fline[q-1] =
                                linebuf[bufnum][fbc[bufnum]].fline[q];
                        linebuf[bufnum][fbc[bufnum]].fline
                            [--linebuf[bufnum][fbc[bufnum]].flen] = '\0';
                        p_cols--;
                        p--;
                    }
                break;
            case EXIT:
            case BREAK:
                memset(linebuf[bufnum][fbc[bufnum]].fline, '\0',
                    (size_t)SCREENMAX);
                linebuf[bufnum][fbc[bufnum]].flen = 0;
                return SS$_NORMAL;
            }
        }
        sprintf(vf_inf.inv.sline, "%s%s", invite[bufnum],
            linebuf[bufnum][fbc[bufnum]].fline);
        CHECKRESULT(smg$label_border(&vdisp.id, outline(vf_inf.inv.sline, -1),
            &vf_inf.inv.pos, &vf_inf.inv.beg, &vf_inf.inv.rend));
        CHECKRESULT(smg$set_physical_cursor(&p_id, &p_rows, &p_cols));
    }
    return SS$_NORMAL;
}


        /* Find a template in text line */
int
find_in_line(void)
{
    int   s, k, lnlin;
    char  search1low, search1up;

    lnlin = line_len - search_len;
    search1low = search_line [0];
    search1up  = TOUPPER(search_line [0]);

    if (direction == FORWARD)
    {
        if (case_sensitive_search == FALSE)
        {
            /* Forward, case insensitive search */
            for (s = find_col; s <= lnlin; s++)
            {
                if ((no_line[s] == search1up) || (no_line[s] == search1low))
                {
                    for (k = 1; k < search_len; k++)
                    {
                        if (TOLOWER(no_line[s + k]) != search_line [k]) break;
                    }
                    if (k == search_len) return s;
                }
            }
        }
        else
        {
            /* Forward, case sensitive search */
            for (s = find_col; s <= lnlin; s++)
            {
                if (no_line[s] == search_line [0])
                {
                    for (k = 1; k < search_len; k++)
                    {
                        if (no_line[s + k] != search_line [k]) break;
                    }
                    if (k == search_len) return s;
                }
            }
        }
    }
    else
    {
        if (find_col > lnlin)
        {
            find_col = lnlin;
        }

        if (case_sensitive_search == FALSE)
        {
            /* Reverse, case insensitive search */
            for (s = find_col; s >= 0; s--)
            {
                if ((no_line[s] == search1up) || (no_line[s] == search1low))
                {
                    for (k = 1; k < search_len; k++)
                    {
                        if (TOLOWER(no_line[s + k]) != search_line [k]) break;
                    }
                    if (k == search_len) return s;
                }
            }
        }
        else
        {
            /* Reverse, case sensitive search */
            for (s = find_col; s >= 0; s--)
            {
                if (no_line[s] == search_line [0])
                {
                    for (k = 1; k < search_len; k++)
                    {
                        if (no_line[s + k] != search_line [k]) break;
                    }
                    if (k == search_len) return s;
                }
            }
        }
    }
    return $_ABNORMAL;
}


        /* Find once again in text mode */
int
txt_nextfind(void)
{
    int findbeg = 0, found = 0;
    int cur_find = 0;
    unsigned short any_key = 0;

    if (rpt == 0)
    {
        if ((vf_inf.end == 0) && (vdisp.cursor.row >= vf_inf.all))
        {
            BEEP;
            return SS$_NORMAL;
        }
        if (find_col < MAXLINE)
        {
            memset(no_line, 'A', (size_t)MAXLINE);
            no_line[MAXLINE] = '\0';
            FBEG(vf_inf.current + vdisp.cursor.row - 1);
            if (sh_fgets(no_line, MAXLINE, SS$_NORMAL) == -1)
                return $_ABNORMAL;
            if (direction == FORWARD)
                find_col++;
            else
                find_col--;
            if ((found = find_in_line()) != $_ABNORMAL)
            {
                find_col = found;
                CHECKRESULT(smg$begin_display_update(&vdisp.id));
                CHECKRESULT(new_page());
                CHECKRESULT(smg$put_chars(&vdisp.id, outline(no_line, found),
                    &vdisp.cursor.row, &1));
                CHECKRESULT(smg$return_cursor_pos(&vdisp.id,
                    &vdisp.cursor.row, &vdisp.cursor.col));
                for (findbeg = 1; findbeg <= vdisp.cursor.col;
                     findbeg += (vdisp.vport.len - 1))
                    CHECKRESULT(smg$change_viewport(&vdisp.id,
                        &vdisp.vport.firstrow, &findbeg));
                vdisp.vport.curcol = vdisp.cursor.col -findbeg +vdisp.vport.len;
                findbeg = 0;
                CHECKRESULT(smg$put_chars(&vdisp.id, outline(&no_line[found],
                    search_len), &vdisp.cursor.row, &vdisp.cursor.col,
                    &findbeg, &rend_find));
                CHECKRESULT(smg$end_display_update(&vdisp.id));
                return SS$_NORMAL;
            }
        }
        if (direction == FORWARD)
        {
            cur_find = vf_inf.current + vdisp.cursor.row;
            if (( (vf_inf.end != $_ABNORMAL) &&
                  ((int)(cur_find - vf_inf.end) >= vdisp.wid) ) ||
                ( (vf_inf.end == 0) &&
                  (vdisp.cursor.row == vf_inf.all - 1) ))
            {
                BEEP;
                return SS$_NORMAL;
            }
        }
        else
        {
            cur_find = vf_inf.current + vdisp.cursor.row - 2;
            if (cur_find < 0)
            {
                BEEP;
                return SS$_NORMAL;
            }
        }
        wn = 0;
        old_percent = -1;
        vdisp.cursor.col++;
    }
    else cur_find = rpt;

    while ( ((direction == FORWARD) && (cur_find < vf_inf.all - 1)) ||
            ((direction != FORWARD) && (cur_find >= 0)) )
    {
        if ((vf_inf.end == $_ABNORMAL) && (cur_find + vdisp.wid >= vf_inf.all))
        {
            rpt = cur_find;
            return SS$_NORMAL;
        }
        FBEG(cur_find);
        switch(working())
        {
        case 0:
            vdisp.cursor.col--;
            return SS$_NORMAL;
        case $_ABNORMAL:
            return $_ABNORMAL;
        }
        if (sh_fgets(no_line, MAXLINE, SS$_NORMAL) == -1) return $_ABNORMAL;
        if (direction == FORWARD)
        {
            find_col = 0;
            found = find_in_line();
            if (found != $_ABNORMAL) break;
            cur_find++;
        }
        else
        {
            find_col = line_len;
            found = find_in_line();
            if (found != $_ABNORMAL) break;
            cur_find--;
        }
    }
    rpt = 0;
    if (found == $_ABNORMAL)
    {
        CHECKRESULT(smg$label_border(&vdisp.id,
            outline(NOTHING_FOUND, -1),
            &vf_inf.inv.pos, &vf_inf.inv.beg, &vf_inf.inv.rend));
        BEEP;
        smg$read_keystroke(&k_id, &any_key);
        CHECKRESULT(smg$set_cursor_mode(&p_id, &c_on));
        vdisp.cursor.col--;
        return SS$_NORMAL;
    }
    find_col = found;
    CHECKRESULT(smg$begin_display_update(&vdisp.id));
    if (cur_find - vf_inf.current >= vdisp.wid)
    {
        if ((vf_inf.end != $_ABNORMAL) && (cur_find > vf_inf.end))
            vf_inf.current = vf_inf.end;
        else if (cur_find <= vdisp.wid/2)
            vf_inf.current = 0;
        else vf_inf.current = cur_find - vdisp.wid/2;
    }
    CHECKRESULT(new_page());
    vdisp.cursor.row = cur_find - vf_inf.current + 1;
    vdisp.cursor.col = 1;
    CHECKRESULT(smg$set_cursor_mode(&p_id, &c_on));
    CHECKRESULT(smg$put_chars(&vdisp.id, outline(no_line, found),
        &vdisp.cursor.row, &vdisp.cursor.col));
    CHECKRESULT(smg$return_cursor_pos(&vdisp.id,
        &vdisp.cursor.row, &vdisp.cursor.col));
    for (findbeg = 1; findbeg <= vdisp.cursor.col;
        findbeg += (vdisp.vport.len - 1))
        CHECKRESULT(smg$change_viewport(&vdisp.id,
            &vdisp.vport.firstrow, &findbeg));
    vdisp.vport.curcol = vdisp.cursor.col - findbeg + vdisp.vport.len;
    findbeg = 0;
    CHECKRESULT(smg$put_chars(&vdisp.id, outline(&no_line[found], search_len),
        &vdisp.cursor.row, &vdisp.cursor.col, &findbeg, &rend_find));
    CHECKRESULT(smg$end_display_update(&vdisp.id));
    return SS$_NORMAL;
}


        /* Find once again in a binary file */
int
bin_nextfind()
{
    int buffer_len, i, j;
    int cur_find, back_find;
    int found = $_ABNORMAL;
    unsigned short any_key = 0;

    if (direction == FORWARD)
    {
        if (rpt == 0)
        {
            cur_find = vdisp.cursor.col - (BINRECLEN + LONGWORDLEN) * 2 - 1;
            if (cur_find < 0) cur_find = 0;
            else if (cur_find > BINRECLEN) cur_find = BINRECLEN;
            cur_find += (vf_inf.current + vdisp.cursor.row - 1) * BINRECLEN;
        }
        else
            cur_find = rpt;

        cur_find++;
        if (cur_find + search_len - 1 > (int) sh_eof())
        {
            if (vf_inf.end == $_ABNORMAL)
                rpt = cur_find;
            else
                BEEP;
            return SS$_NORMAL;
        }
        if (sh_seek(cur_find) == -1) return $_ABNORMAL;
        buffer_len = sh_read (no_line, MAXLINE);
        while (buffer_len >= search_len)
        {
            switch(working())
            {
            case 0:
                return SS$_NORMAL;
            case $_ABNORMAL:
                return $_ABNORMAL;
            }

            if (case_sensitive_search == FALSE)
            {
                /* Forward, case insensitive search */
                for (i = 0; i < buffer_len - search_len; i++)
                {
                    for (j = 0; j < search_len; j++)
                    {
                        if (search_line [j] != TOLOWER(no_line [i + j])) break;
                    }
                    if (j == search_len)
                    {
                        found = cur_find + i;
                        break;
                    }
                }
            }
            else
            {
                /* Forward, case sensitive search */
                for (i = 0; i < buffer_len - search_len; i++)
                {
                    for (j = 0; j < search_len; j++)
                    {
                        if (search_line [j] != no_line [i + j]) break;
                    }
                    if (j == search_len)
                    {
                        found = cur_find + i;
                        break;
                    }
                }
            }

            if (found != $_ABNORMAL) break;

            cur_find += i;
            if (sh_seek(cur_find) == -1)
                buffer_len = 0;
            else
                buffer_len = sh_read (no_line, MAXLINE);

            if ((vf_inf.end == $_ABNORMAL) && (buffer_len < search_len))
            {
                rpt = cur_find;
                return SS$_NORMAL;
            }
        }
    }
    else   /* Do a reverse search */
    {
        cur_find = vdisp.cursor.col - (BINRECLEN + LONGWORDLEN) * 2 - 1;
        if (cur_find < 0) cur_find = 0;
        else if (cur_find > BINRECLEN) cur_find = BINRECLEN;
        cur_find += (vf_inf.current + vdisp.cursor.row - 1) * BINRECLEN;
        cur_find--;

        if (cur_find + search_len > (int) sh_eof())
           cur_find = (int) sh_eof() - search_len + 1;
        if (cur_find < 0)
        {
            BEEP;
            return SS$_NORMAL;
        }
        back_find = cur_find - MAXLINE + search_len;
        if (back_find < 0) back_find = 0;
        if (sh_seek(back_find) == -1) return $_ABNORMAL;
        buffer_len = sh_read (no_line, cur_find - back_find + search_len);
        while (buffer_len >= search_len)
        {
            switch(working())
            {
            case 0:
                return SS$_NORMAL;
            case $_ABNORMAL:
                return $_ABNORMAL;
            }

            if (case_sensitive_search == FALSE)
            {
                /* Reverse, case insensitive search */
                for (i = buffer_len - search_len; i >= 0; i--)
                {
                    for (j = 0; j < search_len; j++)
                    {
                        if (search_line [j] != TOLOWER(no_line [i + j])) break;
                    }
                    if (j == search_len)
                    {
                        found = back_find + i;
                        break;
                    }
                }
            }
            else
            {
                /* Reverse, case sensitive search */
                for (i = buffer_len - search_len; i >= 0; i--)
                {
                    for (j = 0; j < search_len; j++)
                    {
                        if (search_line [j] != no_line [i + j]) break;
                    }
                    if (j == search_len)
                    {
                        found = back_find + i;
                        break;
                    }
                }
            }

            if (found != $_ABNORMAL) break;

            cur_find = back_find - 1;
            back_find = cur_find - MAXLINE + search_len;
            if (back_find < 0) back_find = 0;
            if (sh_seek(back_find) == -1)
                buffer_len = 0;
            else
                buffer_len = sh_read (no_line,
                    cur_find - back_find + search_len);
        }
    }

    if (found == $_ABNORMAL)
    {
        CHECKRESULT(smg$label_border(&vdisp.id,
            outline(NOTHING_FOUND, -1),
            &vf_inf.inv.pos, &vf_inf.inv.beg, &vf_inf.inv.rend));
        BEEP;
        smg$read_keystroke(&k_id, &any_key);
        CHECKRESULT(smg$set_cursor_mode(&p_id, &c_on));
        return SS$_NORMAL;
    }
    else
    {
        find_col = $_ABNORMAL;
        CHECKRESULT(smg$begin_display_update(&vdisp.id));
        cur_find = found / BINRECLEN;
        if (cur_find - vf_inf.current >= vdisp.wid)
        {
            if ((vf_inf.end != $_ABNORMAL) && (cur_find > vf_inf.end))
                vf_inf.current = vf_inf.end;
            else if (cur_find <= vdisp.wid/2)
                vf_inf.current = 0;
            else vf_inf.current = cur_find - vdisp.wid/2;
        }
        CHECKRESULT(new_page());
        vdisp.cursor.row = cur_find - vf_inf.current + 1;
        vdisp.cursor.col = (BINRECLEN + LONGWORDLEN)* 2 + found % BINRECLEN + 1;
        vdisp.vport.curcol = vdisp.cursor.col;
        CHECKRESULT(smg$set_cursor_mode(&p_id, &c_on));
        CHECKRESULT(smg$change_viewport(&vdisp.id,
                    &vdisp.vport.firstrow, &vdisp.vport.firstcol));
        CHECKRESULT(smg$set_cursor_abs(&vdisp.id,
            &vdisp.cursor.row, &vdisp.cursor.col));
        CHECKRESULT(smg$end_display_update(&vdisp.id));
    }

    return SS$_NORMAL;
}


        /* Go to position by line or percentage */
int
go_line()
{
    unsigned int ln = 0, p = 0;
    char *pch = NULL;

    ln = strtoul(linebuf[TOLINE][fbc[TOLINE]].fline, &pch, 0);
    vf_inf.current = 0;
    if (pch == linebuf[TOLINE][fbc[TOLINE]].fline) BEEP
    else
    {
        if (*pch == '%')
        {
            p = (ln * sh_eof())/100;
            if ((vf_inf.end == $_ABNORMAL) && (vf_inf.home[vf_inf.all] < p))
            {
                rpt = $_ABNORMAL;
                return SS$_NORMAL;
            }
            for (ln = 0; (ln < vf_inf.all) && (vf_inf.home[ln] < p); ln++);
            ln++;
        }

        if (ln == 0)
        {
            BEEP;
            ln = 1;
        }

        if ((vf_inf.end == $_ABNORMAL) && (ln + vdisp.wid >= vf_inf.all))
        {
            rpt = $_ABNORMAL;
            return SS$_NORMAL;
        }

        CHECKRESULT(smg$set_cursor_mode(&p_id, &c_on));

        if (ln >= vf_inf.all)
        {
            BEEP;
            ln = vf_inf.all - 1;
        }

        if (ln <= vdisp.wid/2) vdisp.cursor.row = ln;
        else
        {
            vf_inf.current = ln - vdisp.wid/2;
            vdisp.cursor.row = vdisp.wid/2;
            if ((vf_inf.end != $_ABNORMAL) && (vf_inf.current > vf_inf.end))
            {
                vdisp.cursor.row += (vf_inf.current - vf_inf.end);
                vf_inf.current = vf_inf.end;
            }
        }
        CHECKRESULT(new_page());
    }
    rpt = 0;

    return SS$_NORMAL;
}


        /* Key processing */
int
key_processing(unsigned short ch)
{
    int   l = 0;
    int   j;
    char  byte1;
    char  byte2;

        switch(ch)
        {
        case 0:
            return SS$_NORMAL;
        case UP:
            if (vdisp.cursor.row == 1)
            {
                if (vf_inf.current != 0)
                {
                    vf_inf.current--;
                    FBEG(vf_inf.current);
                    l = 1;
                    CHECKRESULT(smg$begin_display_update(&vdisp.id));
                    CHECKRESULT(smg$set_cursor_abs(&vdisp.id,
                        &vdisp.cursor.row, &l));
                    if (sh_fgets(line, MAXLINE, SS$_NORMAL) == -1)
                        return $_ABNORMAL;
                    l = 1;
                    CHECKRESULT(smg$insert_line(&vdisp.id, &l,
                        outline(line, line_len), &pscroll));
                    CHECKRESULT(smg$end_display_update(&vdisp.id));
                }
                else BEEP;
            }
            else vdisp.cursor.row--;
            find_col = vdisp.cursor.col - 1;
            return SS$_NORMAL;
        case DOWN:
            if (vdisp.cursor.row == vdisp.wid)
            {
                if ((vf_inf.end == $_ABNORMAL) &&
                    (vf_inf.current + vdisp.wid + 1 >= vf_inf.all))
                {
                    rpt = $_ABNORMAL;
                    return SS$_NORMAL;
                }
                if (vf_inf.current != vf_inf.end)
                {
                    l = 1;
                    FBEG(vf_inf.current + vdisp.wid);
                    CHECKRESULT(smg$begin_display_update(&vdisp.id));
                    CHECKRESULT(smg$set_cursor_abs(&vdisp.id,
                        &vdisp.cursor.row, &l));
                    if (sh_fgets(line, MAXLINE, SS$_NORMAL) == -1)
                        return $_ABNORMAL;
                    l = vdisp.cursor.row;
                    CHECKRESULT(smg$insert_line(&vdisp.id, &l,
                        outline(line, line_len)));
                    vf_inf.current++;
                    CHECKRESULT(smg$end_display_update(&vdisp.id));
                }
                else BEEP;
            }
            else vdisp.cursor.row++;
            rpt = 0;
            find_col = vdisp.cursor.col - 1;
            return SS$_NORMAL;
        case LEFT:
            if (vdisp.cursor.col > 1)
            {
                if (vdisp.vport.curcol > 1) vdisp.vport.curcol--;
                else CHECKRESULT(smg$scroll_viewport(&vdisp.id, &scrright));
                vdisp.cursor.col--;
                find_col = vdisp.cursor.col - 1;
            }
            else BEEP;
            return SS$_NORMAL;
        case RIGHT:
            if (vdisp.cursor.col < vdisp.len)
            {
                if (vdisp.vport.curcol < vdisp.vport.len) vdisp.vport.curcol++;
                else CHECKRESULT(smg$scroll_viewport(&vdisp.id, &scrleft));
                vdisp.cursor.col++;
                find_col = vdisp.cursor.col - 1;
            }
            else BEEP;
            return SS$_NORMAL;
        case BACKSPACE:
        case 'L':
        case 'l':
            if (vdisp.vport.curcol > 1)
            {
                vdisp.cursor.col -= (vdisp.vport.curcol - 1);
                vdisp.vport.curcol = 1;
                find_col = vdisp.cursor.col - 1;
            }
            else if (vdisp.cursor.col > vdisp.vport.len)
            {
                l = vdisp.vport.len - 1;
                CHECKRESULT(smg$scroll_viewport(&vdisp.id,
                    &scrright, &l));
                vdisp.cursor.col -= (vdisp.vport.len - 1);
                find_col = vdisp.cursor.col - 1;
            }
            else if (vdisp.cursor.col > 1)
            {
                CHECKRESULT(smg$change_viewport(&vdisp.id,
                    &vdisp.vport.firstrow, &vdisp.vport.firstcol));
                vdisp.cursor.col = 1;
                find_col = vdisp.cursor.col - 1;
            }
            else BEEP;
            return SS$_NORMAL;
        case ENDOFLINE:
        case 'R':
        case 'r':
            if (vdisp.vport.curcol < vdisp.vport.len)
            {
                vdisp.cursor.col += (vdisp.vport.len - vdisp.vport.curcol);
                vdisp.vport.curcol = vdisp.vport.len;
                find_col = vdisp.cursor.col - 1;
            }
            else if (vdisp.cursor.col < (vdisp.len - 2 * vdisp.vport.len))
            {
                l = vdisp.vport.len - 1;
                CHECKRESULT(smg$scroll_viewport(&vdisp.id,
                    &scrleft, &l));
                vdisp.cursor.col += (vdisp.vport.len - 1);
                find_col = vdisp.cursor.col - 1;
            }
            else if(vdisp.cursor.col < vdisp.len)
            {
                l = vdisp.len - vdisp.vport.len + 1;
                CHECKRESULT(smg$change_viewport(&vdisp.id,
                    &vdisp.vport.firstrow, &l));
                vdisp.cursor.col = vdisp.len;
                find_col = vdisp.cursor.col - 1;
            }
            else BEEP;
            return SS$_NORMAL;
        case PREV:
            if (vf_inf.current == 0)
            {
               if ((vdisp.cursor.row == 1) && (vdisp.cursor.col == 1)) BEEP
               else
               {
                   vdisp.cursor.row = 1;
                   vdisp.vport.curcol = 1;
                   vdisp.cursor.col = 1;
               }
            }
            else
            {
                if (vf_inf.current >= vdisp.wid)
                    vf_inf.current -= (vdisp.wid - 1);
                else vf_inf.current = 0;
                CHECKRESULT(new_page());
            }
            find_col = 0;
            return SS$_NORMAL;
        case NEXT:
            if ((vf_inf.end == $_ABNORMAL) &&
                (vf_inf.current + 2*vdisp.wid >= vf_inf.all))
                {
                    wn = WORKING_STEP;
                    old_percent = -1;
                    rpt = $_ABNORMAL;
                    find_col = 0;
                    return SS$_NORMAL;
                }
            if (vf_inf.current != vf_inf.end)
            {
                if ((vf_inf.end == $_ABNORMAL) ||
                    (vf_inf.end - vf_inf.current >= vdisp.wid))
                    vf_inf.current += (vdisp.wid - 1);
                else vf_inf.current = vf_inf.end;
                CHECKRESULT(new_page());
            }
            else
            {
               if (vdisp.cursor.row == vdisp.wid) BEEP
               else
               {
                   vdisp.cursor.row = vdisp.wid;
                   vdisp.vport.curcol = 1;
                   vdisp.cursor.col = 1;
               }
            }
            rpt = 0;
            find_col = 0;
            return SS$_NORMAL;
        case 'H':
        case 'h':
            if ((vf_inf.current == 0) &&
                (vdisp.vport.curcol == 1) &&
                (vdisp.cursor.row == 1) &&
                (vdisp.cursor.col == 1)) BEEP
            else
            {
                vdisp.vport.curcol = 1;
                vdisp.cursor.row = vdisp.cursor.col = 1;
                CHECKRESULT(smg$begin_display_update(&vdisp.id));
                if (vf_inf.current != 0)
                {
                    vf_inf.current = 0;
                    CHECKRESULT(new_page());
                }
                CHECKRESULT(smg$change_viewport(&vdisp.id,
                    &vdisp.vport.firstrow, &vdisp.vport.firstcol));
                CHECKRESULT(smg$end_display_update(&vdisp.id));
            }
            find_col = 0;
            return SS$_NORMAL;
        case 'E':
        case 'e':
            if (vf_inf.end == $_ABNORMAL)
            {
                rpt = $_ABNORMAL;
                return SS$_NORMAL;
            }
            if ((vf_inf.current == vf_inf.end) &&
                (vdisp.vport.curcol == 1) &&
                (vdisp.cursor.col == 1) &&
                (vdisp.cursor.row == vdisp.wid)) BEEP
            else
            {
                vdisp.vport.curcol = 1;
                vdisp.cursor.row = vdisp.wid;
                vdisp.cursor.col = 1;
                CHECKRESULT(smg$begin_display_update(&vdisp.id));
                if (vf_inf.current != vf_inf.end)
                {
                    vf_inf.current = vf_inf.end;
                    CHECKRESULT(new_page());
                }
                CHECKRESULT(smg$change_viewport(&vdisp.id,
                    &vdisp.vport.firstrow, &vdisp.vport.firstcol));
                CHECKRESULT(smg$end_display_update(&vdisp.id));
            }
            CHECKRESULT(smg$set_cursor_mode(&p_id, &c_on));
            rpt = 0;
            find_col = 0;
            return SS$_NORMAL;
        case TOGGLE_DIR:
            if (direction == FORWARD)
                direction = REVERSE;
            else
                direction = FORWARD;
            CHECKRESULT(headline());
            return SS$_NORMAL;
        case SET_FORWARD:
        case 'V':
        case 'v':
            direction = FORWARD;
            CHECKRESULT(headline());
            return SS$_NORMAL;
        case SET_REVERSE:
        case 'U':
        case 'u':
            direction = REVERSE;
            CHECKRESULT(headline());
            return SS$_NORMAL;
        case FIND:
        case FIND_CASE:
        case 'F':
        case 'f':
            if (rpt == 0)
            {
                if (vf_inf.bintxt == SS$_NORMAL)
                    /* Normally case insensitive for text mode */
                    if (ch == FIND_CASE)
                        case_sensitive_search = TRUE;
                    else
                        case_sensitive_search = FALSE;
                 else
                    /* Normally case sensitive for binary mode */
                    if (ch == FIND_CASE)
                        case_sensitive_search = FALSE;
                    else
                        case_sensitive_search = TRUE;

                /* Input a search string */
                CHECKRESULT(get_line(TOFIND));

                /* Convert hex sequences to hex bytes */
                j = 0;
                for (l = 0; l < linebuf[TOFIND][fbc[TOFIND]].flen - 3; l++)
                {
                    if ( (linebuf[TOFIND][fbc[TOFIND]].fline [l] == '\\') &&
                         ( (linebuf[TOFIND][fbc[TOFIND]].fline [l+1] == 'X') ||
                           (linebuf[TOFIND][fbc[TOFIND]].fline [l+1] == 'x') ) )
                    {
                        byte1 = linebuf[TOFIND][fbc[TOFIND]].fline [l+2];
                        byte2 = linebuf[TOFIND][fbc[TOFIND]].fline [l+3];
                        if ( ( ((byte1 >= '0') && (byte1 <= '9')) ||
                               ((byte1 >= 'A') && (byte1 <= 'F')) ||
                               ((byte1 >= 'a') && (byte1 <= 'f')) ) &&
                             ( ((byte2 >= '0') && (byte2 <= '9')) ||
                               ((byte2 >= 'A') && (byte2 <= 'F')) ||
                               ((byte2 >= 'a') && (byte2 <= 'f')) ) )
                        {
                            byte1 = byte1 | 0x20;
                            byte2 = byte2 | 0x20;
                            search_line [j] = ( ( ((byte1 & 0x40) == 0)
                                                  ? byte1 - '0'
                                                  : byte1 - 'a' + 10 ) << 4 ) +
                                              ( ((byte2 & 0x40) == 0)
                                                ? byte2 - '0'
                                                : byte2 - 'a' + 10 );
                            j++;
                            l += 3;
                        }
                        else
                        {
                            search_line [j] =
                                linebuf[TOFIND][fbc[TOFIND]].fline [l];
                            j++;
                        }
                    }
                    else
                    {
                        search_line[j] = linebuf[TOFIND][fbc[TOFIND]].fline [l];
                        j++;
                    }
                }

                /* Copy bytes at end that can't be part of hex sequence */
                for (; l < linebuf[TOFIND][fbc[TOFIND]].flen; l++)
                {
                    search_line [j] = linebuf[TOFIND][fbc[TOFIND]].fline [l];
                    j++;
                }
                search_len = j;
                search_line [search_len] = '\0';

                /* Make a lower case version for case insensitive searches */
                if (case_sensitive_search == FALSE)
                {
                    for (l = 0; l < search_len; l++)
                    {
                        search_line [l] = TOLOWER(search_line [l]);
                    }
                }
            }
        case FIND_NEXT:
        case 'N':
        case 'n':
            if (search_len != 0)
                if (vf_inf.bintxt == SS$_NORMAL)
                    return txt_nextfind();
                else
                    return bin_nextfind();
            else
                BEEP;
            return SS$_NORMAL;
        case NEXT_FF:
        case NEXT_ff:
            {
                int  return_status;
                char fline_copy;
                int  flen_copy;

                if (vf_inf.bintxt == SS$_NORMAL)
                {
                    flen_copy = search_len;
                    fline_copy = search_line [0];
                    search_len = 1;
                    search_line [0] = 12;
                    return_status = txt_nextfind();
                    search_len = flen_copy;
                    search_line [0] = fline_copy;
                }
                else
                {
                    flen_copy = search_len;
                    fline_copy = search_line [0];
                    search_len = 1;
                    search_line [0] = 12;
                    return_status = bin_nextfind();
                    search_len = flen_copy;
                    search_line [0] = fline_copy;
                }
                return return_status;
            }
        case 'G':
        case 'g':
            if (rpt == 0) CHECKRESULT(get_line(TOLINE));
            if (linebuf[TOLINE][fbc[TOLINE]].flen != 0) return go_line();
            else BEEP;
            return SS$_NORMAL;
        case 'W':
        case 'w':
            screensize = (screensize == SCREENMAX) ? SCREENMIN : SCREENMAX;
            CHECKRESULT(smg$unpaste_virtual_display(&vdisp.id, &p_id));
            CHECKRESULT(smg$change_pbd_characteristics(&p_id,
                &screensize, &l));
            if (screensize == l)
            {
                vdisp.vport.len = screensize - BORDERFIELD;
                l = vdisp.cursor.col - vdisp.vport.curcol + 1;
                if (l > (vdisp.len - vdisp.vport.len + 1))
                    l = vdisp.len - vdisp.vport.len + 1;
                CHECKRESULT(smg$change_viewport(&vdisp.id, 0, &l,
                    &vdisp.wid, &vdisp.vport.len));
                if (vdisp.vport.curcol > vdisp.vport.len)
                {
                    vdisp.cursor.col -= (vdisp.vport.curcol - vdisp.vport.len);
                    vdisp.vport.curcol = vdisp.vport.len;
                }
                else vdisp.vport.curcol = vdisp.cursor.col - l + 1;
            }
            else BEEP;
            CHECKRESULT(smg$paste_virtual_display(&vdisp.id, &p_id,
                &vdisp.firstrow, &vdisp.firstcol));
            return SS$_NORMAL;
        case 'D':
        case 'd':
            CHECKRESULT(smg$label_border(&vdisp.id, outline(PLEASE_WAIT, -1),
                &vf_inf.inv.pos, &vf_inf.inv.beg, &vf_inf.inv.rend));
            memset(no_line, 'A', (size_t)SCREENMAX);
            line[SCREENMAX] = '\0';
            if (sys$trnlnm (&LNM$M_CASE_BLIND, &tablename, &elnm, &3, item_list)
                != SS$_NORMAL)
                strcpy(line, EDITCOMMAND);
            else
                line [lnmlen] = 0;
            sprintf(no_line, line,
                vdisp.f_spec, vf_inf.current + vdisp.cursor.row);
            return (lib$spawn(outline(no_line, -1)) == SS$_NORMAL)
                ? EXIT_NORMAL : RELOOK;
        case 'P':
        case 'p':
        case hELP:
            memset(line, 'A', (size_t)SCREENMAX);
            line[SCREENMAX] = '\0';
            if (sys$trnlnm (&LNM$M_CASE_BLIND, &tablename, &hlnm, &3, item_list)
                != SS$_NORMAL)
            {
                lnmlen = 0;
                item_list [0].item_code = JPI$_IMAGNAME;
                l = sys$getjpiw (0, 0, 0, item_list, 0, 0, 0);
                item_list [0].item_code = LNM$_STRING;
                line [lnmlen] = 0;
                l = lnmlen;
                while ((l >= 0) && (line [l] != ']')) {
                   line [l] = 0;
                   l--;
                }
                sprintf (no_line, HELPCOMMAND, line);
                strcpy (line, no_line);
            }
            else
                line [lnmlen] = 0;

        case 'S':
        case 's':
            CHECKRESULT(smg$label_border(&vdisp.id, outline(PLEASE_WAIT, -1),
                &vf_inf.inv.pos, &vf_inf.inv.beg, &vf_inf.inv.rend));
            CHECKRESULT(smg$set_physical_cursor(&p_id, &k_bar.firstrow,
                &k_bar.firstcol));
            if ((ch == 'S') || (ch == 's'))
                CHECKRESULT(lib$spawn(0))
            else CHECKRESULT(lib$spawn(outline(line, -1)))
        case REFRESH:
            CHECKRESULT(smg$unpaste_virtual_display(&vdisp.id, &p_id));
            CHECKRESULT(smg$unpaste_virtual_display(&k_bar.id, &p_id));
            CHECKRESULT(smg$erase_pasteboard(&p_id));
            CHECKRESULT(smg$paste_virtual_display(&k_bar.id, &p_id,
                &k_bar.firstrow, &k_bar.firstcol));
            CHECKRESULT(new_page());
            CHECKRESULT(smg$paste_virtual_display(&vdisp.id, &p_id,
                &vdisp.firstrow, &vdisp.firstcol));
            return SS$_NORMAL;
        case 'T':
        case 't':
        case '/':
            vf_inf.bintxt = (vf_inf.bintxt == SS$_NORMAL)
                ? $_ABNORMAL : SS$_NORMAL;
        case REREAD:
            return RELOOK;
        case EXIT:
        case BREAK:
        case 'X':
        case 'x':
            return EXIT_NORMAL;
        case 'Q':
        case 'q':
            return 0;
        }
    return SS$_NORMAL;
}


        /* First acquaintance */
int
get_acquaintance(unsigned short *key)
{
    int key_reply = SS$_NORMAL;
    int mem_all = vdisp.wid + 5;

    memset(line, '\0', (size_t)MAXLINE);
    memset(no_line, '\0', (size_t)MAXLINE);
    rpt         = 0;
    *key        = 0;
    wn          = 0;
    old_percent = -1;

    vf_inf.home     = (unsigned int *)malloc((size_t)(mem_all * sizeof(int)));
    vf_inf.current  = vf_inf.end = vf_inf.all = 0;

    while(vf_inf.all < vdisp.wid)
    {
        vf_inf.home[vf_inf.all] = sh_tell();
        vf_inf.all++;
        if (sh_fgets(line, MAXLINE, SS$_NORMAL) == -1) break;
        CHECKRESULT(smg$put_line(&vdisp.id, outline(line, line_len)));
    }
    CHECKRESULT(smg$paste_virtual_display(&vdisp.id, &p_id,
        &vdisp.firstrow, &vdisp.firstcol));
    CHECKRESULT(smg$paste_virtual_display(&k_bar.id, &p_id,
        &k_bar.firstrow, &k_bar.firstcol));
    if (vf_inf.all < vdisp.wid) return SS$_NORMAL;

    CHECKRESULT(headline());
    vf_inf.end = $_ABNORMAL;

    do
    {
        if (vf_inf.all >= mem_all)
        {
            mem_all *= 2;
            vf_inf.home = (unsigned int *)realloc(vf_inf.home,
                (size_t)(mem_all * sizeof(int)));
        }
        vf_inf.home[vf_inf.all] = sh_tell();

            if (rpt != 0)
            {
                if (rpt == $_ABNORMAL) key_reply = working();
                if (vf_inf.all % READKEY_STEP == 0)
                {
                    if (key_reply == SS$_NORMAL)
                        key_reply = key_processing(*key);
                    else if (key_reply == 0) key_reply = SS$_NORMAL;
                    if (key_reply != SS$_NORMAL) return key_reply;
                    sh_seek(vf_inf.home[vf_inf.all]);
                }
                if (rpt == 0) CHECKRESULT(headline());
            }
            else if ((vf_inf.all % READKEY_STEP == 0) &&
                (smg$read_keystroke(&k_id, key, NULL, &timeout) == SS$_NORMAL))
            {
                wn = 0;
                old_percent = -1;
                key_reply = key_processing(*key);
                if (key_reply != SS$_NORMAL) return key_reply;
                sh_seek(vf_inf.home[vf_inf.all]);
                if (rpt == 0) CHECKRESULT(headline());
            }

        vf_inf.all++;
    }
    while(sh_fgets(line, MAXLINE, $_ABNORMAL) != -1);

    vf_inf.end = vf_inf.all - vdisp.wid - 1;
    return (rpt != 0) ? key_processing(*key) : SS$_NORMAL;
}


        /* Main view routine */
void
view_manager(char file_name_to_view[])
{
    unsigned short svch = 0;
    int num = 0;
    int l;

    if (file_name_to_view[0] == '?')
    {
        putchar('\n');
        memset(line, 'A', (size_t)MAXLINE);
        line[MAXLINE] = '\0';
        if (sys$trnlnm (&LNM$M_CASE_BLIND, &tablename, &hlnm, &3, item_list)
            != SS$_NORMAL)
        {
            lnmlen = 0;
            item_list [0].item_code = JPI$_IMAGNAME;
            l = sys$getjpiw (0, 0, 0, item_list, 0, 0, 0);
            item_list [0].item_code = LNM$_STRING;
            line [lnmlen] = 0;
            l = lnmlen;
            while ((l >= 0) && (line [l] != ']')) {
                line [l] = 0;
                l--;
            }
            sprintf (no_line, HELPCOMMAND, line);
            strcpy (line, no_line);
        }
        else
            line [lnmlen] = 0;
        lib$do_command(outline(line, -1));
        exit(SS$_NORMAL);
    }

    vdisp.f_spec = sh_fname(file_name_to_view);
    if (vdisp.f_spec == NULL)
        printf("\n%%LOOK-F-FFERR, Failed to find file %s.\n", file_name_to_view);
    else
    do
        if (sh_open(vdisp.f_spec) != -1)
        {
            num = (init_screen() == SS$_NORMAL)
                ? get_acquaintance(&svch) : $_ABNORMAL;

            while (num == SS$_NORMAL)
                if (headline() == SS$_NORMAL)
                {
                    if (smg$read_keystroke(&k_id, &svch) == SS$_NORMAL)
                        num = key_processing(svch);
                }
                else num = $_ABNORMAL;

            if (num != $_ABNORMAL)
                switch(free_screen())
                {
                case $_ABNORMAL:    num = $_ABNORMAL;   break;
                case NO_CLOSE:      num = NO_CLOSE;     break;
                }

            switch(num)
            {
            case $_ABNORMAL:
                printf("\n%%LOOK-F-LKERR, Failed to use screen.");
            case 0:
                exit(SS$_NORMAL);
            case RELOOK:
                break;
            case NO_CLOSE:
                printf("\n%%LOOK-F-FCERR, Failed to close file %s.\n",
                    vdisp.f_spec);
            default:
                vdisp.f_spec = sh_fname(NULL);
                vf_inf.bintxt = 0;
            }
        }
        else
        {
            printf("\n%%LOOK-F-FOERR, Failed to open file %s.\n", vdisp.f_spec);
            vdisp.f_spec = sh_fname(NULL);
        }
    while(vdisp.f_spec != NULL);
}


        /* Now, real MAIN */
int
main(int arguments, char *parameter[])
{
    unsigned char buzz[] =
    {
        '\50',
        '\224',
        '\102',
        '\235',
        '\123',
        '\225',
        '\123',
        '\236',
        '\101',
        '\334',
        '\132',
        '\205',
        '\51',
        '\263',
        '\76',
        '\246',
        '\74',
        '\343',
        '\127',
        '\307',
        '\116',
        '\320',
        '\103',
        '\305',
        '\125',
        '\334',
        '\17',
        '\320',
        '\146',
        '\327',
        '\170',
        '\253',
        '\164',
        '\305',
        '\125',
        '\334',
        '\114',
        '\300',
        '\126',
        '\313',
        '\135',
        '\320',
        '\134',
        '\310',
        '\31',
        '\0'
    };
    char a_file_name[FNAMLEN] = "";
    int i = 0;
    char *b_file_name = "Fast file viewer V6.00.";

    for (i = strlen((char *)buzz) - 1; i > 0; i--)
       buzz[i] = (~buzz[i]) ^ buzz[i-1];

    if (arguments > 1)
    {
        printf("%s%s", b_file_name, buzz);
        for (i = 1; i < arguments; i++) view_manager(parameter[i]);
    }
    else
    {
        do
        {
            printf("_File(s) : ");
            gets(a_file_name);
            putchar('\n');
        } while(a_file_name[0] == '\0');

        printf("%s%s", b_file_name, buzz);
        for (b_file_name = strtok(a_file_name, " "); b_file_name != NULL;
            b_file_name = strtok(NULL, " ")) view_manager(b_file_name);
    }

#if 0
    /* To convert string to buzz numbers, use the following */
    for (i = 1; i < strlen((char *) buzz); i++)
       buzz[i] = ~(buzz[i] ^ buzz[i-1]);
    for (i = 0; i <= strlen((char *) buzz); i++)
       printf("   '\\%o',\n", buzz [i]);
#endif
}
